<?php
if (!defined('MEDIAWIKI')) die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * Hook functions required for WikiDB.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2014, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 856 $
 */

WikiDB_RegisterRevision('$Rev: 856 $');

// wfWikiDB_SkinHooks() - for 'SkinTemplateNavigation::Universal' hook
//
// Purpose:
//	To provide the functionality of wfWikiDB_AddActionTabs() in MW 1.18 and above.
//  We simply pass the appropriate bit of the tab array (by reference) to that
//  function, which works pretty-much the same in both contexts.
	function wfWikiDB_SkinHooks(&$SkinTemplate, &$Links) {
		$ActionTabs =& $Links['namespaces'];
		wfWikiDB_AddActionTabs($ActionTabs);
		return true;
	}

// wfWikiDB_AddActionTabs() - for 'SkinTemplateContentActions' hook
//
// Purpose:
// 	If viewing a table namespace, add a 'data' tab after the subject-page tab.
//	This tab displays the number of records and is a redlink if there are none.
//	It provides the 'viewdata' action for the page in question (always the subject
//  page even if viewing the talk page).
//  Note that in MW >= 1.18, the wfWikiDB_SkinHooks() function is called instead,
//  but it simply acts as a wrapper to this function, which is the same for both
//  situations, except for the method it uses to work out which tab the new tab
//  should follow.
//
// TODO: Each namespace is allowed its own default text for the subject page tab, so
//		we should respect that in a way that allows us to be clear that this is the
//		definition tab... maybe we just need sensible defaults?
// TODO: Should we remove the 'edit' tab when action='viewdata'?  Any other
//		 non-applicable tabs which should be removed?
	function wfWikiDB_AddActionTabs(&$ContentActions) {
		global $wgTitle;
		
	// Get the requested action.
	// We use this to check whether we are on the 'viewdata' tab or not, in order
	// to ensure the appropriate styling is applied.
		$action = WikiDB_GetAction();

		$Title = $wgTitle;

	// The data tab is the same for the main page and the talk page, so if we're on
	// the talk page then we need to change $Title to point to the subject page
	// instead.
		if ($Title->isTalkPage())
			$Title = $Title->getSubjectPage();

	// If we're in a DB namespace, add a new 'data' tab.
		if (WikiDB_IsValidTable($Title)) {
		// Get URL for 'data' tab.
			$Link = $Title->getLocalURL('action=viewdata');

		// Count records for this table, to display the number of rows in the text
		// for the tab.
			$Table = new WikiDB_Table($Title);
			$Count = $Table->CountRows();

		// If we are on the 'data' tab, then the class should be marked as selected.
			if ($action == "viewdata")
				$Class = "selected";
			else
				$Class = "";

		// If no records exist, colour tab 'red' in the same way as a link to a
		// missing page, indicating that there is no data. If there are records it
		// will have the standard link style for existing pages.
			if ($Count == 0)
				$Class .= " new";

		// Add new tab to the array.  Because we want to add it after the 'content'
		// tab, rather than at the end, we need to rebuild the array.
			$NewActions = array();
			foreach ($ContentActions as $Key => $Value) {
				$NewActions[$Key] = $Value;

			// If we are viewing the data page, make sure that no other action is
			// selected.  (If we don't do this then the subject page tab will also
			// be selected.)
				if ($action == "viewdata") {
					$NewActions[$Key]['class'] =
						preg_replace('/selected/', "", $NewActions[$Key]['class']);
				}

			// If this is the subject tab (which always begins "nstab-", followed by
			// the namespace name) then change its label to the text
			// "table definition" and add the 'data' tab as the next tab.
				if (substr($Key, 0, 6) == "nstab-"
					|| (isset($Value['context']) && $Value['context'] == "subject"))
				{
				// Change label of subject page tab.
					$NewActions[$Key]['text'] = wfMsg('nstab-wikidb-tabledef');

				// Add the 'data' tab between the subject-page tab and the talk-page
				// tab.
					$NewActions['viewdata'] = array(
						'class' => $Class,
						'text' => wfMsg('nstab-wikidb-tabledata', $Count),
						'href' => $Link,
					// These two are only present in MW 1.18 and above, but there
					// is no harm in adding them for lower versions so we don't need
					// to do a version check here.
						'primary' => 1,
						'context' => "viewdata",
					);
				}
			}
			$ContentActions = $NewActions;
		}
		return true;
	}

// wfWikiDB_SpecialActions() - for 'UnknownAction' hook
//
// Purpose:
//	To handle the 'viewdata' action.  This action only applies in a table namespace,
//	and is ignored in other namespaces.
//	If this action is requested, we output all table data for the specified table
//	in the standard WikiDB format.
//
// TODO: Should we raise an error if the action happens in a non-table namespace, or
//		is the current behaviour OK?
// TODO: Currently you get the same output whether the requested page is the subject
//		page or the talk page.  Is this the best behaviour?  One potential problem
//		is that this means there are two URLs for the same page content - does this
// 		matter?  If so, there are two other options: (1) disallow 'viewdata' for
//		talk pages.  Not a problem when using the interface, but makes URL wrangling
//		a big more fussy (2) perform a hard redirect so that the URL changes to the
//		subject page, when accessed from the talk page.
	function wfWikiDB_SpecialActions($action, $article) {
		global $wgOut, $wgRequest;

	// If the requested action is not 'viewdata' then don't do anything.
	// We return true, indicating that MediaWiki should carry on as normal.
		if ($action != "viewdata")
			return true;

	// The data is the same for the main page and the talk page, so if we're on the
	// talk page then we need to change $Title to point to the subject page instead.
	// In practice this won't happen through the interface, as the tab always links
	// to the subject page, but it may occur through URL wrangling.
		$Title = $article->getTitle();
		if ($Title->isTalkPage())
			$Title = $Title->getSubjectPage();

	// If this is not a table namespace then ignore the action.
	// We return true, indicating that MediaWiki should carry on as normal.
	// This will result in an 'invalid action' error message.
		if (!WikiDB_IsValidTable($Title))
			return true;

	// Get range parameters from the URL
		$ItemsPerPage = $wgRequest->getIntOrNull('limit');
		if (!$ItemsPerPage)
			$ItemsPerPage = 50;
		$Offset = $wgRequest->getInt('offset');

	// Set page title.
		$wgOut->setPageTitle(wfMsg('wikidb-tabledata-heading', 
								   $Title->getFullText()));

	// Setup the table object.
		$Table = new WikiDB_Table($Title->getFullText());

	// If the table is an alias, it holds no data directly.  We instead add a link
	// to the table it is an alias of.
	// TODO: Or should we automatically redirect?
		if ($Table->IsAlias()) {
			$Dest = $Table->GetDataDestination();
			$Title = $Dest->GetTitle();
			$URL = $Title->getInternalURL("action=viewdata");
			$wgOut->addWikiText("<big>'''This table is a redirect.'''\n\n"
								. "All data for this table ends up in [" 
								. $URL . " " . $Dest->GetFullName() . "]</big>");
			return false;
		}

	// Get the table data.
		$TotalRows = $Table->CountRows();
		$Data = $Table->GetRows($Offset, $ItemsPerPage);

	// If no data found, report the fact.
		if (count($Data) == 0)
			$wgOut->addWikiText(wfMsg('wikidb-nodata'));

	// Otherwise, output the table data in the standard format.
		else {
		// Output pagination controls.
			$wgOut->addHTML('<p>' 
							. wfShowingResults($Offset, $ItemsPerPage) 
							. '</p>');

			$LastRowOnPage = $Offset + $ItemsPerPage;
			if ($LastRowOnPage >= $TotalRows)
				$IsLastPage = true;
			else
				$IsLastPage = false;

			$Output = wfViewPrevNext(
									 $Offset,
									 $ItemsPerPage, 
									 $Title->getFullText(),
									 "action=viewdata", 
									 $IsLastPage
									);

			$wgOut->addHTML( '<p>' . $Output . '</p>' );

		// Output the data.
			$wgOut->addWikiText($Table->FormatTableData($Data, true));
		}

	// Return false to tell MediaWiki that we have successfully generated the page
	// contents and to skip processing of other hooks.
		return false;
	}

// wfWikiDB_ArticleSaveComplete() - for 'ArticleSave' hook.
//
// Purpose:
//	To update the WikiDB tables when an article is saved.  It is essentially a
//  wrapper to WikiDB::PageUpdated()
//
// Notes:
// * This hook is called when an article is created, edit and rolled-back.
// * The whole article text is passed, even if the edit was to a section.
// * There are additional arguments to this function, but I have omitted them as
//	 we don't use them.
	function wfWikiDB_ArticleSaveComplete(&$article, &$user, $text) {
		WikiDB::PageUpdated($article, $text);
		return true;
	}

// wfWikiDB_ArticleUndelete() - for 'ArticleUndelete' hook
//
// Purpose:
//	To update the WikiDB tables when an article is undeleted.  It is essentially a
//	wrapper to WikiDB::PageUpdated().
//
// Note: This hook was added In MW 1.9.1.  For versions prior to that,
// wfWikiDB_SpecialPageExecuteAfterPage() is called instead of this.
	function wfWikiDB_ArticleUndelete($title, $create) {
		WikiDB::PageUpdated($title);
		return true;
	}

// wfWikiDB_SpecialPageExecuteAfterPage() - for 'SpecialPageExecuteAfterPage' hook
//
// Purpose:
//	To update the WikiDB tables when an article is undeleted.  It is essentially a
//  wrapper to WikiDB::PageUpdated().
//
// Note: In MW 1.9.1, the ArticleUndelete hook was added, so
// wfWikiDB_ArticleUndelete() is called instead of this.
	function wfWikiDB_SpecialPageExecuteAfterPage($specialPage, $par, $funct) {
		global $wgOut;

	// If this is the 'undelete' page
		if ($specialPage->getName() == "Undelete") {
		// If this is an undelete action (instead of 'view deleted' or any other
		// functionality)
			if (isset($_POST['restore'])) {
				$Page = $_POST['target'];

				$SuccessMsg = wfMsg("undeletedtext", $Page);
				$SuccessMsg = $wgOut->parse($SuccessMsg, true);

			// Check if the delete was a success (i.e. whether the success message
			// is in the page output.
				if (strpos($wgOut->getHTML(), $SuccessMsg) !== false) {
					$Title = Title::newFromText($Page);
					WikiDB::PageUpdated($Title);
				}
			}
		}

		return true;
	}

// wfWikiDB_ParserBeforeStrip() - for 'ParserBeforeStrip' hook.
//
// Purpose:
//	To render pages in the table namespace (to format defintions correctly, and to
//	add the summary box at the top).
//  This used to hook ParserAfterStrip, but as of MW 1.14 this is deprecated and
//  behaves identically to ParserBeforeStrip, so we now use that instead to ensure
//  we get the same behaviour in all MW versions.
	function wfWikiDB_ParserBeforeStrip(&$parser, &$text, &$strip_state) {
		global $wgpWikiDBParsingForSave;

	// Static flag to stop us re-running this code if any subsequent code invokes
	// the parser for any reason.  We set this to true once we know this is a
	// page in the table namespace, so that it isn't run a second time (resulting
	// in duplicate headers etc. being added to the page.
	// This can happen via WikiDB if, for example, a <data> tag is added to a
	// table definition page.  It is also possible that other extensions might
	// trigger a reparse, so even if WikiDB is modified to avoid this, the check
	// should remain.
	// We only skip if we are not parsing for a save - save operations should always
	// be allowed through so that the DB is updated properly!
		static $ParsingComplete = false;
		if ($ParsingComplete && !$wgpWikiDBParsingForSave)
			return true;

		$Title = $parser->mTitle;

	// We only apply the formatting to pages in a table namespace (and not their
	// talk pages, either), so skip rest of function if not in table namespace.
		if (!WikiDB_IsValidTable($Title))
			return true;

	// If the page is being parsed for a save operation, call the function
	// that does the DB update and exit.
		if ($wgpWikiDBParsingForSave) {
			WikiDB::pAddTableDefinition($Title, $text);
			return true;
		}

	// We only want to parse the page content - not any system messages, etc.
		$ParserOptions = $parser->getOptions();
	// According to IAlex (see IRC logs for #mediawiki, @19:06 on 5th Jan 2013),
	// we can tell if this is a system message by checking if either
	// getInterfaceMessage() is true or getEnableLimitReport() is false.
	// Otherwise it is page content.
	// If the functions are available, we use these method to check whether it is
	// page content or not.
		if (method_exists($ParserOptions, "getInterfaceMessage")) {
			if ($ParserOptions->getInterfaceMessage()
				|| (method_exists($ParserOptions, "getEnableLimitReport")
					&& !$ParserOptions->getEnableLimitReport()))
			{
				return true;
			}
		}
	// However this function was not present in earlier MW versions (I have checked
	// 1.17 where it exists, and 1.6 where it doesn't.  I don't know at what stage
	// it was introduced).
	// For earlier versions, we check the value of getTidy() which seems to return
	// true for the page output, but not for interface code, so we can detect whether
	// this is a page or not by checking that variable.
	// Note that this method, intuitively, seems wrong.  However it is what we have
	// been using for many years, and I have not yet had any bug reports related to
	// this - and I would expect to, as it completely breaks the table namespace - 
	// so I think it is fairly safe.
		elseif (!$ParserOptions->getTidy()) {
			return true;
		}

	// Don't add to data if we're editing the page.
		switch (WikiDB_GetAction()) {
		// These are OK
			case "purge":
			case "submit":
			case "view":
				break;
		// The rest are bad
			default:
				return true;
		}

		$DB =& wfGetDB(DB_SLAVE);
	
	// Parse the text to get the field definitions (including any extra text) ready
	// for outputting.
	// Previously we pulled this from the table definition, but in practice this
	// code is called before our ArticleSaveComplete hook, therefore we end up
	// displaying the previous table definition instead of the new one (requiring
	// a page purge in order to update properly).  Now we always use the text as
	// passed into the function, even though this results in an extra parse.
//		$Table = new WikiDB_Table($Title);
//		$Fields = $Table->GetFields();
		$Fields = WikiDB_Parser::ParseTableDef($text);

	// Format the page contents using the WikiDB parser function.
	// We include a __NOEDITSECTION__ tag at the start, so that the 'edit'
	// buttons don't display, as our code adds a couple of headings and this
	// means that they don't work properly.
	// TODO: Can we fix this somehow?
		$text = "__NOEDITSECTION__\n" . WikiDB_Table::FormatTableDef($Fields);

	// If previewing, don't show the summary at the top.
	// REMOVED - WE PROBABLY SHOULD SHOW THIS...
//			if (WikiDB_GetAction() == "submit")
//				return true;

	// If showing the summary, use a heading for the source.
		$text = "== " . wfMsg('wikidb-tabledef-sourceheading') . " ==\n" . $text;

		if (count($Fields) > 0) {
			$DefinedFields = array();
			$Aliases = array();

			$TableData = array();
			$HasComments = false;
			$HasForeignKeys = false;
			$AliasesHaveComments = false;

			foreach ($Fields as $Field) {
			// Skip over any non-field lines.
				if (!is_array($Field))
					continue;

				if (isset($Field['Alias'])) {
					$Aliases[$Field['FieldName']] = $Field['Alias'];
					if (isset($Field['Comment']))
						$AliasesHaveComments = true;
				}
				else {
					$TableRow = array();

					$TableRow['FieldName'] = $Field['FieldName'];

					if (isset($Field['Type'])
						&& WikiDB_TypeHandler::TypeDefined($Field['Type']))
					{
						$TableRow['Type'] = $Field['Type'];
						if (isset($Field['Options'])
							&& count($Field['Options']) > 0)
						{
							$TableRow['Type'] .=
									" (" . implode(",", $Field['Options']) . ")";
						}
					}
					else
						$TableRow['Type'] = "&nbsp;";

					if (isset($Field['ForeignTable'])) {
						$ForeignTable = new WikiDB_Table($Field['ForeignTable'],
												$Title->getNamespace());
						$TableRow['ForeignKey'] =
										"[[" . $ForeignTable->GetFullName() . "|"
										. $ForeignTable->GetFullName() . "#"
										. $Field['ForeignField']
										. "]]";
						$HasForeignKeys = true;
					}
					else
						$TableRow['ForeignKey'] = "&nbsp;";

					if (isset($Field['Comment'])) {
						$TableRow['Comment'] = $Field['Comment'];
						$HasComments = true;
					}
					else
						$TableRow['Comment'] = "&nbsp;";
					$TableData[] = $TableRow;
					$DefinedFields[] = $Field['FieldName'];
				}
			}

			$Colspan = 2;
			$Definition = '<table border="1" cellpadding="5" cellspacing="2">'
						. "\n\t" . '<tr>'
						. '<th>' . wfMsg('wikidb-tabledef-field') . '</th>'
						. '<th>' . wfMsg('wikidb-tabledef-type') . '</th>';
			if ($HasForeignKeys) {
				$Colspan++;
				$Definition .= '<th>' . wfMsg('wikidb-tabledef-foreignkey') 
							 . '</th>';
			}
			if ($HasComments) {
				$Colspan++;
				$Definition .= '<th>' . wfMsg('wikidb-tabledef-comment') . '</th>';
			}
			$Definition .= '</tr>' . "\n";

			if (count($TableData) > 0) {
				foreach ($TableData as $TableRow) {
					$Definition .= "\t<tr>\n";
					$Definition .= "\t\t<td>" . $TableRow['FieldName']
								 . "</td>\n";
					$Definition .= "\t\t<td>" . $TableRow['Type'] . "</td>\n";
					if ($HasForeignKeys) {
						$Definition .= "\t\t<td>" . $TableRow['ForeignKey']
									 . "</td>\n";
					}
					if ($HasComments) {
						$Definition .= "\t\t<td>" . $TableRow['Comment']
									 . "</td>\n";
					}
					$Definition .= "\t</tr>\n";
				}
			}
			else {
				$Definition .= "\t" . '<tr><td colspan="' . $Colspan
							 . '"><i>' . wfMsg('wikidb-tabledef-nofields') 
							 . '</i></td></tr>' . "\n";
			}

			$Definition .= "</table>\n";

			if (count($Aliases) > 0) {
				$Definition .= "\n=== " . wfMsg('wikidb-tabledef-aliasesheading') 
							 . " ===\n";
				$Definition .= '<table border="1" cellpadding="5" '
							 . 'cellspacing="2">' . "\n";
				$Definition .= "\t<tr><th>" . wfMsg('wikidb-tabledef-field') 
							 . "</th><th>" . wfMsg('wikidb-tabledef-alias') 
							 . "</th>";
				if ($AliasesHaveComments) {
					$Definition .= '<th>' . wfMsg('wikidb-tabledef-comment') 
								 . '</th>';
				}
				$Definition .= "</tr>\n";
				foreach ($Fields as $Field) {
					if (is_array($Field) && isset($Field['Alias'])) {
						$Definition .= "\t<tr>\n";
						$Definition .= "\t\t<td>" . $Field['FieldName']
									 . "</td>\n";
						$Definition .= "\t\t<td>" . $Field['Alias'] . "</td>\n";
						if ($AliasesHaveComments) {
							if (isset($Field['Comment'])) {
								$Definition .= "\t\t<td>" . $Field['Comment']
											 . "</td>\n";
							}
							else
								$Definition .= "\t\t<td>&nbsp;</td>\n";
						}
						$Definition .= "\t</tr>\n";
					}
				}
				$Definition .= "</table>\n";
			}
			$text = "== " . wfMsg('wikidb-tabledef-definitionheading') . " ==\n" 
				  . $Definition . "\n" . $text;
		}

	// Set the $ParsingComplete flag to stop any reparse.
		$ParsingComplete = true;
		return true;
	}

// wfWikiDB_ArticleDeleteComplete() - for 'ArticleDeleteComplete' hook.
//
// Purpose:
//	To tidy up our database when a page is deleted.
	function wfWikiDB_ArticleDeleteComplete(&$article, &$user, $reason) {
		WikiDB::PageDeleted($article);
		return true;
	}

// fnWikiDB_TitleMoveComplete() - for 'TitleMoveComplete' hook
//
// Purpose:
//	To update the WikiDB tables with the new article title when an article is moved.
//
// Previously this function worked by manipulating the data in the DB so that we
// avoided having to do a full re-parse.  However I kept encountering scenarios that
// I hadn't catered for, and which would leave the DB containing incorrect data.
// I have therefore re-written the function to simply re-parse both the old and new
// locations, which should guarantee that the DB is fully and correctly updated, at 
// the (assumed) expense of execution speed.
	function wfWikiDB_TitleMoveComplete(&$title, &$newtitle, &$user, $oldid, $newid)
	{
	// Reparse the original location, to update the WikiDB tables for that page.
	// Note that this code assumes that WikiDB::PageUpdated() will correctly
	// handle situations where the title doesn't exist (because in later versions
	// of MediaWiki leaving behind a redirect is optional).  This situation has not
	// been fully tested.  If it causes problems then WikiDB::PageUpdated(), 
	// is the place that needs fixing, not this function.
		WikiDB::PageUpdated($title);

	// Need to reset the article ID for the new title, because currently it is not 
	// considered to exist because of stale data in the object.  Calling this 
	// function causes a proper refresh, so we get the correct page content.
		$newtitle->resetArticleID(0);

	// Reparse the new location, to update the WikiDB tables for the page.
		WikiDB::PageUpdated($newtitle);

		return true;
	}

// wfWikiDB_AddCSS()
// Adds the extension's CSS file to the skin's CSS output.  The output is generated
// as part of the <head> section, which is not ideal, but I've only tested it on 
// v1.6.10 - perhaps it has been changed to a remote inclusion in later versions of 
// MW.  Either way, I don't think we have much choice, as we can't guarantee that
// the extension's directory is web-accessible.  This method ensures the CSS works
// out-of-the-box, rather than requiring any additional steps to embed the CSS.
	function wfWikiDB_AddCSS(&$Output) {
	// Get the contents of our WikiDB.css file.
		$NewCSS = file_get_contents(dirname(__FILE__) . "/../WikiDB.css");

	// As of 1.16 the CSS data is automatically wrapped in CDATA tags when
	// necessary.  For earlier versions, we do this ourselves.  The method for
	// applying this has been copied from Html::inlineStyle().  As the class
	// was added in 1.16 we can simply check for its existence, rather than doing
	// a version check.
		if (class_exists("Html") && preg_match( '/[<&]/', $NewCSS)) {
			$NewCSS = "/*<![CDATA[*/\n" . $NewCSS . "\n/*]]>*/";
		}

	// Add the CSS to the output.  We include a new-line so that multiple additions
	// are separated out clearly.
		$Output .= "\n" . $NewCSS;

	// Return true, to allow other hooks to run.
		return true;
	}

// Testing function.  Add this to a hook and it will print out the value
// of the arguments passed into the hook (actually, it displays the arguments of the
// calling function, whether it is a hook or not).
// Passing true into the function will cause the script to die() after the arguments
// have been displayed, otherwise it will return false (allowing you to use
//   return DisplayHookInput();
// to cancel the hook having displayed the arguments.
// As this is purely for debugging, there is no point using wfMsg() for the strings.
	function DisplayHookInput($KillScript = false) {
		$bt = debug_backtrace();

		print("<pre>\n");
		if (isset($bt[1]['args']))
			print_r($bt[1]['args']);
		else
			print("No arguments passed to function.\n");
		print("</pre>\n");

		if ($KillScript)
			die();
		else
			return false;
	}
